/* sendip.c
	Sends specified ip packets
*/

/* Command line options:
	./sendip <hostname> -p <type> -d <data> <options>
	
	<hostname> is the name of the host to send the packet to, defaults to
	localhost
	<type> is the packet type: currently IP, UDP, TCP or ICMP, defaults to IP
	<data> is the data payload of the packet as a string of hexadecimal
	numbers, defaults to 10 zero bytes

	For all options, specify the option, followed by its value.  Any option
	can be left out, in which case the default value is used.  For all options,
	if you specify "r", sendip will generate a random value.

	OPTION     DESCRIPTION     DEFAULT
	Socket options (values are 0 or 1)
	-sd        Debug           0
	-sr        Don't route     0 (ie do route)
	-sb        Allow broadcast 0
	
	IP options:
	-is        Source IP       127.0.0.1
	-id        Destination IP  Correct
	-ih        Header length   Correct
	-iy        Type of service 0 (see RFC791 et al)
	-il        Length          Correct
	-ii        Identification  Random
	-ifr       Reserved flag   0
	-ifd       Don't fragment  0
	-ifm       More fragments  0
	-if        Frag offset     0
	-it        TTL             255
	-ip        Protocol        Correct for type
	-ic        Checksum        Correct
	-io        Options - not yet implemented
	
	ICMP options:
	-ct        Type            8 (See RFC792)
	-cd        Code            0 (See RFC792)
	-cc        Checksum        Correct
	
	UDP options:
	-us        Source port      0
	-ud        Destination port 0
	-ul        Length          Correct
	-uc        Checksum        Correct
	
	TCP options:
	-ts        Source port      0
	-td		  Destination port 0
	-tn        Sequence number Random
	-ta        Acknowledgement 0
	-tt        Data offset     Correct
	-tr        Reserved        0
	-tfu       Set urg bit     0 unless -tu is specified
	-tfa       Set ack bit     0 unless -ta is specified
	-tfp       Set psh bit     0
	-tfr       Set rst bit     0
	-tfs       Set syn bit     1
	-tff       Set fin bit     0
	-tw        Window          65535
	-tc        Checksum        Correct
	-tu        Urgent pointer  0
	-to        Options - not yet implemented

   RIP options: (RIP is transmitted using UDP so UDP options also apply)
   You can have up to 25 RIP entries or 1 authenticate (must be first) and up
   to 24 RIP entries in a valid RIP packet. The well-known UDP port for RIP-1
   and RIP-2 is 520. Messages intended for another router's process should be
   sent to UDP 520. Routing update messages should be sent from UDP 520.
   Unsolicited routing update messages (which is what you will be doing, I
   would guess ;-) should have both their origin and destination UDP ports 520.

Command:     Name:        Subcategory:   Default:      Valid range:
  -rv        RIP version                 2               1 or 2
  -rc        RIP command                 1               1-request, 2-response
For each of the RIP entry fields, type - if you want the default to be used
  -re        RIP entry    Address family AF_INET (2)     AF_INET6 (10) not
                                                         supported (generally)
                                                         for RIP-1
                          Route tag      0               not valid for RIP-1
                          IPv4 address   0.0.0.0         any IPv4 address
                          Subnet mask    255.255.255.0   any netmask - not
                                                         valid for RIP-1
                          Next hop       0.0.0.0         not valid for RIP-1
                          Metric         16              1-16
  -ra        authenticate                Brave new world any string - not valid
                                                         for RIP-1 (quote the
                                                         string!!)
             The authenticate option uses THE FIRST of the 25 availble routing
             entries. To distinguish this type of entry from a routing entry,
             the address family is 0xFFFF (and the authentication tag is 2) 

  RIP actions:
  -rd        RIP default request. Ask for the routers' entire routing tables.

*/

#define __USE_BSD   /* GLIBC */
#define _BSD_SOURCE /* LIBC5 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <rpc/types.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include "protocols.h"

/* IP_MAXPACKET is not defined on pre-glibc boxes so define it in case */
#ifndef IP_MAXPACKET
#define IP_MAXPACKET 65535
#endif

/* MAXOOSTNAMELEN is not defined on BSD boxes */
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

void do_sockopt(int s, char *opt, char *val);
void usage_exit(const char *err);

const char *progname;

/* print usage message and exit */
void usage_exit(const char *err) {
	fprintf(stderr,"Usage: %s <hostname> <options>\n",progname);
	fprintf(stderr,err);
	fprintf(stderr,"See sendip.c or the man page for more information\n");
	exit(1);
}

/* handle socket options */
void do_sockopt(int s, char *opt, char *val) {
	const int zero = 0;
	const int one = 1;
	int r;

	if(*val != '0' && *val != '1' && *val != 'r') {
		close(s);
		usage_exit("Value for socket option must be 0, 1 or r\n");
	}
	if(*val=='r') {
		*val=rand()&1;
	} else {
		*val-=0x30;
	}

	switch(opt[2]) {
	case 'd':
	case 'D':
		if(*val)
			r = setsockopt(s, SOL_SOCKET, SO_DEBUG, &one, sizeof(one));
		else
			r = setsockopt(s, SOL_SOCKET, SO_DEBUG, &zero, sizeof(zero));
		if(r)
			perror("Warning: setsockopt(SO_DEBUG)");
		break;
	case 'r':
	case 'R':
		if(*val)
			r = setsockopt(s, SOL_SOCKET, SO_DONTROUTE, &one, sizeof(one));
		else
			r = setsockopt(s, SOL_SOCKET, SO_DONTROUTE, &zero, sizeof(zero));
		if(r)
			perror("Warning: setsockopt(SO_DONTROUTE)");
		break;
	case 'b':
	case 'B':
		if(*val)
			r = setsockopt(s, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one));
		else
			r = setsockopt(s, SOL_SOCKET, SO_BROADCAST, &zero, sizeof(zero));
		if(r)
			perror("Warning: setsockopt(SO_DONTROUTE)");
		break;
	default:
		close(s);
		usage_exit("Unknown socket option\n");
		break;
	}
}	


/* Main program code */
int main(int argc, char **argv) {
	/* socket stuff */
	int s;                            /* socket for sending   */

	/* hostname stuff */
	char hnamebuf[MAXHOSTNAMELEN];    /* buffer for hostname  */
	char *hostname = NULL;            /* target hostname      */
	struct hostent *hp;               /* hostname             */
	struct sockaddr whereto;          /* destination          */
	struct sockaddr_in *to = (struct sockaddr_in *)&whereto;
	                                  /*             address  */

	/* packet stuff */
	unsigned char outpack[IP_MAXPACKET];
                                     /* actual output packet */
	struct ip *iph=(struct ip *)outpack;
	                                  /* IP packet header     */
	unsigned char *outp;
	struct rip_options rippack;       /* RIP data packet      */

	struct icmp *icp=NULL;            /* ICMP packet header   */
	struct tcphdr *tcp=NULL;          /* TCP packet header    */
	struct udphdr *udp=NULL;          /* UDP packet header    */
	int rip = 0;                      /* Are we doing RIP     */
	char *data = NULL;                /* packet data argument */

	int sent;                         /* number of bytes sent */

	srand(time(NULL));

	/* Variable initializations */
	progname=argv[0];
   memset(to, 0, sizeof(struct sockaddr));
	memset(outpack, 0, IP_MAXPACKET);
	memset(hnamebuf, 0, MAXHOSTNAMELEN);
	outp = outpack+sizeof(struct ip);
	ip_initialize(iph);

	/* Process command line options - first make sure we have a host name */
	if(argc < 2 || argv[1][0]=='-') usage_exit("No hostname specified\n");
	 
	/* handle hostname - need to end up with whereto initialized */
   if(inet_aton(argv[1], &to->sin_addr)) { /* Esay case - we are given an IP */
		to->sin_family = AF_INET;
      hostname = argv[1];
   } else {                                /* Got a host name */
      hp = gethostbyname(argv[1]);
      if(hp == NULL) {
			perror("gethostbyname");
			return 2;
      }
      to->sin_family = hp->h_addrtype;
      if (hp->h_length > (int)sizeof(to->sin_addr)) { /* make sure it'll fit */
         hp->h_length = sizeof(to->sin_addr);
      }
      memcpy(&to->sin_addr, hp->h_addr, hp->h_length); /* copy address and */
      (void)strncpy(hnamebuf, hp->h_name, sizeof(hnamebuf) - 1); /* name   */
      hostname = hnamebuf;
   }
	
	/* Now we can open the socket before handling the rest of the command line
		options */
	if((s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("Couldn't open RAW socket");
		return 2;
	}

	/* Now process the rest of the command line options */
	argc-=2;
	argv+=2;

	while(argc > 1) {
		if(argv[0][0]!='-') {
			close(s);
			usage_exit("Option not starting with a -\n");
		}
		switch(argv[0][1]) {
		case 'c':  /* icmp options */
			if(icp)
				do_icmpopt(icp,argv[0],argv[1]);
			else {
				close(s);
				usage_exit("Can not specify -c* without -p ICMP\n");
			}
			break;
		case 'd':  /* packet data */
			data=argv[1];
			break;
		case 'i':  /* ip options */
			do_ipopt(iph,argv[0],argv[1]);
			break;
		case 'p':  /* packet type */
			if(icp || udp || tcp) {
				close(s);
				usage_exit("Can't specify multiple protocols\n");
			}
			switch(argv[1][0]) {
			case 'I':
			case 'i':
				if(strcasecmp(argv[1],"icmp")) {
					close(s);
					usage_exit("Protocol unknown\n");
				}
				icp=(struct icmp *)outp;
				icmp_initialize(icp);
				outp+=4; /*sizoef(struct icmphdr) includes too much gubbins */;
				break;
			case 'U':
			case 'u':
				if(strcasecmp(argv[1],"UDP")) {
					close(s);
					usage_exit("Protocol unknown\n");
				}
				udp=(struct udphdr *)outp;
				udp_initialize(udp);
				outp+=sizeof(struct udphdr);
				break;
			case 'T':
			case 't':
				if(strcasecmp(argv[1],"TCP")) {
					close(s);
					usage_exit("Protocol unknown\n");
				}
				tcp=(struct tcphdr *)outp;
				tcp_initialize(tcp);
				outp+=sizeof(struct tcphdr);
				break;
			default:
				close(s);
				usage_exit("Protocol unknown\n");
				break;
			}
			break;  /* end of packet type */
		case 'r':  /* RIP options -- this is a UDP protocol */
			if( !udp)
			{
				close(s);
				usage_exit("Can not specify -r* without -p UDP\n");
			}
			else /* specify the rip options */
			{
				rip = 1;
				do_ripopt( &rippack, argv[0], &argv[1]);
				if( (strcmp( argv[0], "-re") == 0) || (strcmp( argv[0], "-ra") == 0))
				{
					rip_finalize( &rippack, outp);
					outp += 24;
				}
				if( strcmp( argv[0], "-re") == 0)
				{
					argc -= 5;
					argv += 5;
				}
			}
			break;
		case 's':  /* socket options */
			do_sockopt(s,argv[0],argv[1]);
			break;
		case 't':  /* tcp options */
			if(tcp)
				do_tcpopt(tcp,argv[0],argv[1]);
			else {
				close(s);
				usage_exit("Can not specify -t* without -p TCP\n");
			}
			break;
		case 'u':  /* udp options */
			if(udp)
				do_udpopt(udp,argv[0],argv[1]);
			else {
				close(s);
				usage_exit("Can not specify -u* without -p UDP\n");
			}
			break;
		default:  /* something else */
			close(s);
			fprintf( stderr, "%s\n", argv[0]);
			usage_exit("Unknown option\n");
			break;
		}
		argc-=2; argv+=2;
	}
	if(argc==1) {
		close(s);
		usage_exit("Argument needs a parameter\n");
	}

	if(!rip) {
		/* finalize packet for sending - first, copy data */
		/* Add data from outp - this code is foul */
		if(data) {
			for(;outp-outpack<IP_MAXPACKET && *data;data++, outp++) {
				if(isdigit(*data)) {
					*outp=(*data)-'0';
				} else if('A' <= *data && *data <= 'F') {
					*outp=(*data)-'A'+0x0a;
				} else if('a' <= *data && *data <= 'f') {
					*outp=(*data)-'a'+0x0a;
				}
				(*outp)*=0x10;
				data++;
				if(*data) {
					if(isdigit(*data)) {
						*outp+=(*data)-'0';
					} else if('A' <= *data && *data <= 'F') {
						*outp+=(*data)-'A'+0x0a;
					} else if('a' <= *data && *data <= 'f') {
						*outp+=(*data)-'a'+0x0a;
					}
				} else data--; /* force loop exit */
			}
		} else {  /* no data - just add 10 bytes to size */
			outp+=10;
			if(outp-outpack > IP_MAXPACKET) 
				outp = outpack+IP_MAXPACKET;
		}
	} /* !rip */

	/* then do any firther finalization (eg checksums) */
	ip_set_addr(iph, to);
	if(icp) icmp_finalize(iph, icp, outp-outpack);
	if(udp) udp_finalize(iph, udp, outp-outpack);
	if(tcp) tcp_finalize(iph, tcp, outp-outpack);
	ip_finalize(iph, outp-outpack, to);

	/* Send the packet */
	sent = sendto(s, (char *)outpack, outp-outpack, 0, &whereto,
					  sizeof(struct sockaddr));
	if(sent == outp-outpack) {
		fprintf(stderr,"Sent %d bytes to %s\n",sent,hostname);
	} else {
      if (sent < 0)
			perror("sendto");
		else
			fprintf(stderr,"Only sent %d of %d bytes to %s\n",sent,outp-outpack,
					  hostname);
   }
	close(s);
	return 0;
}
